@@ -1,15 +1,18 @@
-# $Id: LibXSLT.pm,v 1.38 2002/09/04 06:58:53 matt Exp $
-
+# $Id: LibXSLT.pm,v 1.48 2004/03/01 18:42:25 matt Exp $
 package XML::LibXSLT;
 
 use strict;
-use vars qw($VERSION @ISA);
+use vars qw($VERSION @ISA $USE_LIBXML_DATA_TYPES);
 
-use XML::LibXML 1.49;
+use XML::LibXML 1.57;
+use XML::LibXML::Literal;
+use XML::LibXML::Boolean;
+use XML::LibXML::Number;
+use XML::LibXML::NodeList;
 
 require Exporter;
 
-$VERSION = "1.52";
+$VERSION = "1.57";
 
 require DynaLoader;
 
@@ -17,6 +20,8 @@ require DynaLoader;
 
 bootstrap XML::LibXSLT $VERSION;
 
+$USE_LIBXML_DATA_TYPES = 0;
+
 sub new {
     my $class = shift;
     my %options = @_;
@@ -24,17 +29,53 @@ sub new {
     return $self;
 }
 
+# ido - perl dispatcher
+sub perl_dispatcher {
+    my $func = shift;
+    my @params = @_;
+    my @perlParams;
+    
+    my $i = 0;
+    while (@params) {
+        my $type = shift(@params);
+        if ($type eq 'XML::LibXML::Literal' or 
+            $type eq 'XML::LibXML::Number' or
+            $type eq 'XML::LibXML::Boolean')
+        {
+            my $val = shift(@params);
+            unshift(@perlParams, $USE_LIBXML_DATA_TYPES ? $type->new($val) : $val);
+        }
+        elsif ($type eq 'XML::LibXML::NodeList') {
+            my $node_count = shift(@params);
+            my @nodes = splice(@params, 0, $node_count);
+            # warn($_->getName) for @nodes;
+            unshift(@perlParams, $type->new(@nodes));
+        }
+    }
+    
+    $func = "main::$func" unless ref($func) || $func =~ /(.+)::/;
+    no strict 'refs';
+    my $res = $func->(@perlParams);
+    return $res;
+}
+
+
 sub xpath_to_string {
     my @results;
     while (@_) {
         my $value = shift(@_); $value = '' unless defined $value;
         push @results, $value;
-        next if @results % 2;
-        if ($value =~ s/'/', "'", '/g) {
-	    $results[-1] = "concat('$value')";
+        if (@results % 2) {
+            # key
+            $results[-1] =~ s/:/_/g; # XSLT doesn't like names with colons
         }
         else {
-            $results[-1] = "'$results[-1]'";
+            if ($value =~ s/'/', "'", '/g) {
+	        $results[-1] = "concat('$value')";
+            }
+            else {
+                $results[-1] = "'$results[-1]'";
+            }
         }
     }
     return @results;
@@ -86,24 +127,30 @@ sub close_callback {
 sub parse_stylesheet {
     my $self = shift;
     if (!ref($self) || !$self->{XML_LIBXSLT_MATCH}) {
+        #warn "callbacks: $XML::LibXML::match_cb $XML::LibXML::open_cb $XML::LibXML::read_cb $XML::LibXML::close_cb";
         return $self->_parse_stylesheet(@_);
     }
     local $XML::LibXML::match_cb = $self->{XML_LIBXSLT_MATCH};
     local $XML::LibXML::open_cb = $self->{XML_LIBXSLT_OPEN};
     local $XML::LibXML::read_cb = $self->{XML_LIBXSLT_READ};
     local $XML::LibXML::close_cb = $self->{XML_LIBXSLT_CLOSE};
+
+    #warn "localised callbacks: $XML::LibXML::match_cb $XML::LibXML::open_cb $XML::LibXML::read_cb $XML::LibXML::close_cb";
     $self->_parse_stylesheet(@_);
 }
 
 sub parse_stylesheet_file {
     my $self = shift;
     if (!ref($self) || !$self->{XML_LIBXSLT_MATCH}) {
+        #warn "callbacks: $XML::LibXML::match_cb $XML::LibXML::open_cb $XML::LibXML::read_cb $XML::LibXML::close_cb";
         return $self->_parse_stylesheet_file(@_);
     }
     local $XML::LibXML::match_cb = $self->{XML_LIBXSLT_MATCH};
     local $XML::LibXML::open_cb = $self->{XML_LIBXSLT_OPEN};
     local $XML::LibXML::read_cb = $self->{XML_LIBXSLT_READ};
     local $XML::LibXML::close_cb = $self->{XML_LIBXSLT_CLOSE};
+
+    #warn "localised callbacks: $XML::LibXML::match_cb $XML::LibXML::open_cb $XML::LibXML::read_cb $XML::LibXML::close_cb";
     $self->_parse_stylesheet_file(@_);
 }
 
@@ -191,10 +238,15 @@ current date and time as a string:
   </xsl:template>
   </xsl:stylesheet>
 
-If you pass parameters to your extension function they are all down
-converted into strings at runtime - there is no internal access to
-nodelists. The return from your function is also just a plain string,
-there is no support for returning a nodelist.
+Parameters can be in whatever format you like. If you pass in a nodelist
+it will be a XML::LibXML::NodeList object in your perl code, but ordinary
+values (strings, numbers and booleans) will be ordinary perl scalars. If
+you wish them to be C<XML::LibXML::Literal>, C<XML::LibXML::Number> and
+C<XML::LibXML::Number> values respectively then set the variable
+C<$XML::LibXSLT::USE_LIBXML_DATA_TYPES> to a true value. Return values can
+be a nodelist or a plain value - the code will just do the right thing.
+But only a single return value is supported (a list is not converted to
+a nodelist).
 
 =head1 API
 
@@ -212,6 +264,11 @@ parse_stylesheet in an eval{} block to trap this.
 
 Exactly the same as the above, but parses the given filename directly.
 
+=head2 Input Callbacks
+
+To define XML::LibXSLT specific input callbacks, reuse the XML::LibXML
+input callback API as described in L<XML::LibXML::Document(3)/"Input Callbacks">.
+
 =head1 XML::LibXSLT::Stylesheet
 
 The main API is on the stylesheet, though it is fairly minimal.
